from girder_worker.plugins.vtk import vtkrow_to_dict
import vtk


def process_node(vtknode, node):
    num_children = input.GetNumberOfChildren(vtknode)
    if num_children > 0:
        node["children"] = []
    for c in range(num_children):
        vtkchild = input.GetChild(vtknode, c)
        v = vtkrow_to_dict(input.GetVertexData(), vtkchild)
        edge = vtk.vtkGraphEdge()
        input.GetInEdge(vtkchild, 0, edge)
        vtkparentedge = edge.GetId()
        e = vtkrow_to_dict(input.GetEdgeData(), vtkparentedge)
        n = {"edge_data": e, "node_data": v}
        process_node(vtkchild, n)
        node["children"].append(n)

node_fields = []
for c in range(input.GetVertexData().GetNumberOfArrays()):
    node_fields.append(input.GetVertexData().GetAbstractArray(c).GetName())

edge_fields = []
for c in range(input.GetEdgeData().GetNumberOfArrays()):
    edge_fields.append(input.GetEdgeData().GetAbstractArray(c).GetName())

vtkroot = input.GetRoot()

output = {
    "node_fields": node_fields,
    "edge_fields": edge_fields,
    "node_data": vtkrow_to_dict(input.GetVertexData(), vtkroot)
}

process_node(vtkroot, output)



import contextlib
import functools
import imp
import os
import requests
import girder_worker
import select
import shutil
import subprocess
import sys
import tempfile
import time
import traceback


class JobStatus(object):
    INACTIVE = 0
    QUEUED = 1
    RUNNING = 2
    SUCCESS = 3
    ERROR = 4
    CANCELED = 5

    FETCHING_INPUT = 820
    CONVERTING_INPUT = 821
    CONVERTING_OUTPUT = 822
    PUSHING_OUTPUT = 823


class TerminalColor(object):
    """
    Provides a set of values that can be used to color text in the terminal.
    """
    ERROR = '\033[1;91m'
    SUCCESS = '\033[32m'
    WARNING = '\033[1;33m'
    INFO = '\033[35m'
    ENDC = '\033[0m'

    @staticmethod
    def _color(tag, text):
        return ''.join((tag, text, TerminalColor.ENDC))

    @staticmethod
    def error(text):
        return TerminalColor._color(TerminalColor.ERROR, text)

    @staticmethod
    def success(text):
        return TerminalColor._color(TerminalColor.SUCCESS, text)

    @staticmethod
    def warning(text):
        return TerminalColor._color(TerminalColor.WARNING, text)

    @staticmethod
    def info(text):
        return TerminalColor._color(TerminalColor.INFO, text)


class JobManager(object):
    """
    This class is a context manager that can be used to write log messages to
    Girder by capturing stdout/stderr printed within the context and sending
    them in a rate-limited manner to Girder. This is not threadsafe since it
    changes the global values of sys.stdout/sys.stderr.

    It also exposes utilities for updating other job fields such as progress
    and status.
    """
    def __init__(self, logPrint, url, method=None, headers=None, interval=0.5):
        """
        :param on: Whether print messages should be logged to the job log.
        :type on: bool
        :param url: The job update URL.
        :param method: The HTTP method to use when updating the job.
        :param headers: Optional HTTP header dict
        :param interval: Minimum time interval at which to send log updates
        back to Girder over HTTP (seconds).
        :type interval: int or float
        """
        self.logPrint = logPrint
        self.method = method or 'PUT'
        self.url = url
        self.headers = headers or {}
        self.interval = interval
        self.status = None

        self._last = time.time()
        self._buf = ''
        self._progressTotal = None
        self._progressCurrent = None
        self._progressMessage = None

        if logPrint:
            self._pipes = sys.stdout, sys.stderr
            sys.stdout, sys.stderr = self, self

    def __enter__(self):
        return self

    def __exit__(self, excType, excValue, tb):
        """
        When the context is exited, if we have a non-empty buffer, we flush
        the remaining contents and restore sys.stdout and sys.stderr to their
        previous values. We also set the job status to ERROR if we exited the
        context with an exception, or SUCCESS otherwise.
        """
        if excType:
            msg = '%s: %s\n%s' % (
                excType, excValue, ''.join(traceback.format_tb(tb)))

            self.write(msg)
            self.updateStatus(JobStatus.ERROR)
        else:
            self.updateStatus(JobStatus.SUCCESS)

        self._flush()
        self._redirectPipes(False)

    def _redirectPipes(self, redirect):
        if self.logPrint:
            if redirect:
                sys.stdout, sys.stderr = self, self
            else:
                sys.stdout, sys.stderr = self._pipes

    def _flush(self):
        """
        If there are contents in the buffer, send them up to the server. If the
        buffer is empty, this is a no-op.
        """
        if not self.url:
            return

        if len(self._buf) or self._progressTotal or self._progressMessage or \
                self._progressCurrent is not None:
            self._redirectPipes(False)

            requests.request(
                self.method.upper(), self.url, allow_redirects=True,
                headers=self.headers, data={
                    'log': self._buf,
                    'progressTotal': self._progressTotal,
                    'progressCurrent': self._progressCurrent,
                    'progressMessage': self._progressMessage
                })
            self._buf = ''

            self._redirectPipes(True)

    def write(self, message, forceFlush=False):
        """
        Append a message to the log for this job. If logPrint is enabled, this
        will be called whenever stdout or stderr is printed to. Otherwise it
        can be called manually and will still perform rate-limited flushing to
        the server.

        :param message: The message to append to the job log.
        :type message: str
        :param forceFlush: Whether to force the write of this event to the
            server. Useful if you don't expect another update for some time.
        :type forceFlush: bool
        """
        if self.logPrint:
            self._pipes[0].write(message)

        if type(message) == unicode:
            message = message.encode('utf8')

        self._buf += message
        if forceFlush or time.time() - self._last > self.interval:
            self._flush()
            self._last = time.time()

    def updateStatus(self, status):
        """
        Update the status field of a job.

        :param status: The status to set on the job.
        :type status: JobStatus
        """
        if not self.url or status is None or status == self.status:
            return

        self.status = status
        self._redirectPipes(False)
        requests.request(self.method.upper(), self.url, headers=self.headers,
                         data={'status': status}, allow_redirects=True)
        self._redirectPipes(True)

    def updateProgress(self, total=None, current=None, message=None,
                       forceFlush=False):
        """
        Update the progress information about a job.

        :param total: The total progress value, or None to leave the same.
        :type total: int, float, or None
        :param current: The current progress value, or None to leave the same.
        :type current: int, float, or None
        :param message: Progress message to set, or None to leave the same.
        :type message: str or None
        :param forceFlush: Whether to force the write of this event to the
            server. Useful if you don't expect another update for some time.
        :type forceFlush: bool
        """
        if total is not None:
            self._progressTotal = total
        if current is not None:
            self._progressCurrent = current
        if message is not None:
            self._progressMessage = message

        if forceFlush or time.time() - self._last > self.interval:
            self._flush()
            self._last = time.time()


def toposort(data):
    """
    General-purpose topological sort function. Dependencies are expressed as a
    dictionary whose keys are items and whose values are a set of dependent
    items. Output is a list of sets in topological order. This is a generator
    function that returns a sequence of sets in topological order.

    :param data: The dependency information.
    :type data: dict
    :returns: Yields a list of sorted sets representing the sorted order.
    """
    if not data:
        return

    # Ignore self dependencies.
    for k, v in data.items():
        v.discard(k)

    # Find all items that don't depend on anything.
    extra = functools.reduce(
        set.union, data.itervalues()) - set(data.iterkeys())
    # Add empty dependences where needed
    data.update({item: set() for item in extra})

    # Perform the toposort.
    while True:
        ordered = set(item for item, dep in data.iteritems() if not dep)
        if not ordered:
            break
        yield ordered
        data = {item: (dep - ordered)
                for item, dep in data.iteritems() if item not in ordered}
    # Detect any cycles in the dependency graph.
    if data:
        raise Exception('Cyclic dependencies detected:\n%s' % '\n'.join(
                        repr(x) for x in data.iteritems()))


@contextlib.contextmanager
def tmpdir(cleanup=True):
    # Make the temp dir underneath tmp_root config setting
    root = os.path.abspath(girder_worker.config.get('girder_worker', 'tmp_root'))
    try:
        os.makedirs(root)
    except OSError:
        if not os.path.isdir(root):
            raise
    path = tempfile.mkdtemp(dir=root)

    try:
        yield path
    finally:
        # Cleanup the temp dir
        if cleanup and os.path.isdir(path):
            shutil.rmtree(path)


def with_tmpdir(fn):
    """
    This function is provided as a convenience to allow use as a decorator of
    a function rather than using "with tmpdir()" around the whole function
    body. It passes the generated temp dir path into the function as the
    special kwarg "_tempdir".
    """
    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        if '_tempdir' in kwargs:
            return fn(*args, **kwargs)

        cleanup = kwargs.get('cleanup', True)
        with tmpdir(cleanup=cleanup) as tempdir:
            kwargs['_tempdir'] = tempdir
            return fn(*args, **kwargs)
    return wrapped


class PluginNotFoundException(Exception):
    pass


def load_plugins(plugins, paths, ignore_errors=False, quiet=False):
    """
    Enable a list of plugins.

    :param plugins: The plugins to enable.
    :type plugins: list or tuple of str
    :param paths: Plugin search paths.
    :type paths: list or tuple of str
    :param ignore_errors: If a plugin fails to load, this determines whether to
        raise the exception or simply print an error and keep going.
    :type ignore_errors: bool
    :param quiet: Optionally suppress printing status messages.
    :type quiet: bool
    :return: Set of plugins that were loaded successfully.
    :rtype: set
    """
    loaded = set()
    for plugin in plugins:
        try:
            load_plugin(plugin, paths)
            loaded.add(plugin)
            if not quiet:
                print(TerminalColor.success('Loaded plugin "%s"' % plugin))
        except Exception:
            print(TerminalColor.error(
                'ERROR: Failed to load plugin "%s":' % plugin))
            if ignore_errors:
                traceback.print_exc()
            else:
                raise

    return loaded


def load_plugin(name, paths):
    """
    Enable a plugin for the worker runtime.

    :param name: The name of the plugin to load, which is also the name of its
        containing directory.
    :type name: str
    :param paths: Plugin search paths.
    :type paths: list or tuple of str
    """
    if 'girder_worker.plugins' not in sys.modules:
        module = imp.new_module('girder_worker.plugins')
        girder_worker.plugins = module
        sys.modules['girder_worker.plugins'] = module

    for path in paths:
        plugin_dir = os.path.join(path, name)
        if os.path.isdir(plugin_dir):
            moduleName = 'girder_worker.plugins.' + name

            if moduleName not in sys.modules:
                fp, pathname, description = imp.find_module(name, [path])
                module = imp.load_module(moduleName, fp, pathname, description)
                setattr(girder_worker.plugins, name, module)

                if hasattr(module, 'load'):
                    module.load({
                        'plugin_dir': plugin_dir
                    })
            break
    else:
        raise PluginNotFoundException(
            'Plugin "%s" not found. Looked in: \n   %s\n' % (
                name, '\n   '.join(paths)))


def run_process(command, outputs, print_stdout, print_stderr):
    p = subprocess.Popen(args=command, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    fds = [p.stdout, p.stderr]
    while True:
        ready = select.select(fds, (), fds, 1)[0]

        if p.stdout in ready:
            buf = os.read(p.stdout.fileno(), 1024)
            if buf:
                if print_stdout:
                    sys.stdout.write(buf)
                else:
                    outputs['_stdout']['script_data'] += buf
            else:
                fds.remove(p.stdout)
        if p.stderr in ready:
            buf = os.read(p.stderr.fileno(), 1024)
            if buf:
                if print_stderr:
                    sys.stderr.write(buf)
                else:
                    outputs['_stderr']['script_data'] += buf
            else:
                fds.remove(p.stderr)
        if (not fds or not ready) and p.poll() is not None:
            break
        elif not fds and p.poll() is None:
            p.wait()

    return p



import json
import os
import girder_worker.events
import girder_worker.format
import girder_worker.io
from girder_worker.format import converter_path, get_validator_analysis, Validator
from ConfigParser import ConfigParser
from executors.python import run as python_run
from executors.workflow import run as workflow_run
from . import utils


PACKAGE_DIR = os.path.dirname(os.path.abspath(__file__))

# Read the configuration files
_cfgs = ('worker.dist.cfg', 'worker.local.cfg')
config = ConfigParser()
config.read([os.path.join(PACKAGE_DIR, f) for f in _cfgs])

# Maps task modes to their implementation
_task_map = {}


def register_executor(name, fn):
    """
    Register a new executor in the girder_worker runtime. This is used to map the
    "mode" field of a task to a function that will execute the task.

    :param name: The value of the mode field that maps to the given function.
    :type name: str
    :param fn: The implementing function.
    :type fn: function
    """
    _task_map[name] = fn


def unregister_executor(name):
    """
    Unregister an executor from the map.

    :param name: The name of the executor to unregister.
    :type name: str
    """
    del _task_map[name]


# Register the core executors that are always enabled.
register_executor('python', python_run)
register_executor('workflow', workflow_run)

# Load plugins that are enabled in the config file
_plugins = os.environ.get('WORKER_PLUGINS_ENABLED',
                          config.get('girder_worker', 'plugins_enabled'))
_plugins = [p.strip() for p in _plugins.split(',') if p.strip()]
_paths = os.environ.get('WORKER_PLUGIN_LOAD_PATH',
                        config.get('girder_worker', 'plugin_load_path')).split(':')
_paths = [p for p in _paths if p.strip()]
_paths.append(os.path.join(PACKAGE_DIR, 'plugins'))
utils.load_plugins(_plugins, _paths)


def load(task_file):
    """
    Load a task JSON into memory, resolving any ``"script_uri"`` fields
    by replacing it with a ``"script"`` field containing the contents pointed
    to by ``"script_uri"`` (see :py:mod:`girder_worker.uri` for URI formats). A
    ``script_fetch_mode`` field may also be set

    :param analysis_file: The path to the JSON file to load.
    :returns: The analysis as a dictionary.
    """

    with open(task_file) as f:
        task = json.load(f)

    if "script" not in task and task.get("mode") != "workflow":
        prevdir = os.getcwd()
        parent = os.path.dirname(task_file)
        if parent != "":
            os.chdir(os.path.dirname(task_file))
        task["script"] = girder_worker.io.fetch({
            "url": task["script_uri"]
        })
        os.chdir(prevdir)

    return task


def isvalid(type, binding, fetch=True, **kwargs):
    """
    Determine whether a data binding is of the appropriate type and format.

    :param type: The expected type specifier string of the binding.
    :param binding: A binding dict of the form
        ``{"format": format, "data", data}``, where ``format`` is the format
        specifier string, and ``data`` is the raw data to test.
        The dict may also be of the form
        ``{"format": format, "uri", uri}``, where ``uri`` is the location of
        the data (see :py:mod:`girder_worker.uri` for URI formats).
    :param fetch: Whether to do an initial data fetch before conversion
        (default ``True``).
    :returns: ``True`` if the binding matches the type and format,
        ``False`` otherwise.
    """
    analysis = get_validator_analysis(Validator(type, binding["format"]))
    outputs = girder_worker.run(analysis, {"input": binding}, auto_convert=False,
                                validate=False, fetch=fetch, **kwargs)
    return outputs["output"]["data"]


def convert(type, input, output, fetch=True, status=None, **kwargs):
    """
    Convert data from one format to another.

    :param type: The type specifier string of the input data.
    :param input: A binding dict of the form
        ``{"format": format, "data", data}``, where ``format`` is the format
        specifier string, and ``data`` is the raw data to convert.
        The dict may also be of the form
        ``{"format": format, "uri", uri}``, where ``uri`` is the location of
        the data (see :py:mod:`girder_worker.uri` for URI formats).
    :param output: A binding of the form
        ``{"format": format}``, where ``format`` is the format
        specifier string to convert the data to.
        The binding may also be in the form
        ``{"format": format, "uri", uri}``, where ``uri`` specifies
        where to place the converted data.
    :param fetch: Whether to do an initial data fetch before conversion
        (default ``True``).
    :returns: The output binding
        dict with an additional field ``"data"`` containing the converted data.
        If ``"uri"`` is present in the output binding, instead saves the data
        to the specified URI and
        returns the output binding unchanged.
    """
    if fetch:
        input["data"] = girder_worker.io.fetch(input, **kwargs)

    if input["format"] == output["format"]:
        data = input["data"]
    else:
        data_descriptor = input
        for c in converter_path(Validator(type, input['format']),
                                Validator(type, output['format'])):
            result = girder_worker.run(
                c, {"input": data_descriptor}, auto_convert=False,
                status=status, **kwargs)
            data_descriptor = result["output"]
        data = data_descriptor["data"]

    if status == utils.JobStatus.CONVERTING_OUTPUT:
        job_mgr = kwargs.get('_job_manager')
        _job_status(job_mgr, utils.JobStatus.PUSHING_OUTPUT)
    girder_worker.io.push(data, output, **kwargs)
    return output


def _job_status(mgr, status):
    if mgr:
        mgr.updateStatus(status)


@utils.with_tmpdir  # noqa
def run(task, inputs=None, outputs=None, auto_convert=True, validate=True,
        fetch=True, status=None, **kwargs):
    """
    Run a task with the specified I/O bindings.

    :param task: Specification of the task to run.
    :type task: dict
    :param inputs: Specification of how input objects should be fetched
        into the runtime environment of this task.
    :type inputs: dict
    :param outputs: Speficiation of what should be done with outputs
        of this task.
    :type outputs: dict
    :param auto_convert: If ``True`` (the default), perform format conversions
        on inputs and outputs with :py:func:`convert` if they do not
        match the formats specified in the input and output bindings.
        If ``False``, an expection is raised for input or output bindings
        do not match the formats specified in the analysis.
    :param validate: If ``True`` (the default), perform input and output
        validation with :py:func:`isvalid` to ensure input bindings are in the
        appropriate format and outputs generated by the script are
        formatted correctly. This guards against dirty input as well as
        buggy scripts that do not produce the correct type of output. An
        invalid input or output will raise an exception. If ``False``, perform
        no validation.
    :param write_script: If ``True`` task scripts will be written to file before
        being passed to ``exec``. This improves interactive debugging with
        tools such as ``pdb`` at the cost of additional file I/O. Note that
        when passed to run *all* tasks will be written to file including
        validation and conversion tasks.
    :param fetch: If ``True`` will perform a fetch on the input before
        running the task (default ``True``).
    :param status: Job status to update to during execution of this task.
    :type status: girder_worker.utils.JobStatus
    :returns: A dictionary of the form ``name: binding`` where ``name`` is
        the name of the output and ``binding`` is an output binding of the form
        ``{"format": format, "data": data}``. If the `outputs` param
        is specified, the formats of these bindings will match those given in
        `outputs`. Additionally, ``"data"`` may be absent if an output URI
        was provided. Instead, those outputs will be saved to that URI and
        the output binding will contain the location in the ``"uri"`` field.
    """
    def extractId(spec):
        return spec["id"] if "id" in spec else spec["name"]

    if inputs is None:
        inputs = {}

    task_inputs = {extractId(d): d for d in task.get("inputs", ())}
    task_outputs = {extractId(d): d for d in task.get("outputs", ())}
    mode = task.get("mode", "python")

    if mode not in _task_map:
        raise Exception("Invalid mode: %s" % mode)

    job_mgr = kwargs.get('_job_manager')

    info = {
        'task': task,
        'task_inputs': task_inputs,
        'task_outputs': task_outputs,
        'mode': mode,
        'inputs': inputs,
        'outputs': outputs,
        'auto_convert': auto_convert,
        'validate': validate,
        'kwargs': kwargs
    }
    girder_worker.events.trigger('run.before', info)

    try:
        # If some inputs are not there, fill in with defaults
        for name, task_input in task_inputs.iteritems():
            if name not in inputs:
                if "default" in task_input:
                    inputs[name] = task_input["default"]
                else:
                    raise Exception("Required input '%s' not provided." % name)

        for name, d in inputs.iteritems():
            task_input = task_inputs[name]

            # Fetch the input
            if fetch:
                if status == utils.JobStatus.RUNNING and 'data' not in d:
                    _job_status(job_mgr, utils.JobStatus.FETCHING_INPUT)
                d["data"] = girder_worker.io.fetch(
                    d, **dict({'task_input': task_input}, **kwargs))

            # Validate the input
            if validate and not girder_worker.isvalid(
                    task_input["type"], d,
                    **dict({'task_input': task_input, 'fetch': False}, **kwargs)):
                raise Exception(
                    "Input %s (Python type %s) is not in the expected type (%s) "
                    "and format (%s)." % (
                        name, type(d["data"]), task_input["type"], d["format"])
                    )

            # Convert data
            if auto_convert:
                converted = girder_worker.convert(
                    task_input["type"], d, {"format": task_input["format"]},
                    status=utils.JobStatus.CONVERTING_INPUT,
                    **dict({'task_input': task_input, 'fetch': False}, **kwargs))
                d["script_data"] = converted["data"]
            elif (d.get("format", task_input.get("format")) ==
                  task_input.get("format")):
                d["script_data"] = d["data"]
            else:
                raise Exception("Expected exact format match but '%s != %s'." % (
                    d["format"], task_input["format"])
                )

        # Make sure all outputs are there
        if outputs is None:
            outputs = {}

        for name, task_output in task_outputs.iteritems():
            if name not in outputs:
                outputs[name] = {"format": task_output["format"]}

        # Set the appropriate job status flag
        _job_status(job_mgr, status)

        # Actually run the task for the given mode
        _task_map[mode](task=task, inputs=inputs, outputs=outputs,
                        task_inputs=task_inputs, task_outputs=task_outputs,
                        auto_convert=auto_convert, validate=validate, **kwargs)

        for name, task_output in task_outputs.iteritems():
            d = outputs[name]
            script_output = {"data": d["script_data"],
                             "format": task_output["format"]}

            # Validate the output
            if validate and not girder_worker.isvalid(
                    task_output["type"], script_output,
                    **dict({'task_output': task_output}, **kwargs)):
                raise Exception(
                    "Output %s (%s) is not in the expected type (%s) and format "
                    " (%s)." % (
                        name, type(script_output["data"]), task_output["type"],
                        d["format"])
                    )

            # We should consider refactoring the logic below, reasoning about
            # the paths through this code is difficult, since this logic is
            # entered by "run", "isvalid", and "convert".
            if auto_convert:
                outputs[name] = girder_worker.convert(
                    task_output["type"], script_output, d,
                    status=utils.JobStatus.CONVERTING_OUTPUT,
                    **dict({'task_output': task_output}, **kwargs))
            elif d["format"] == task_output["format"]:
                data = d["script_data"]

                if status == utils.JobStatus.RUNNING:
                    _job_status(job_mgr, utils.JobStatus.PUSHING_OUTPUT)
                girder_worker.io.push(
                    data, d, **dict({'task_output': task_output}, **kwargs))
            else:
                raise Exception("Expected exact format match but %s != %s.'" % (
                    d["format"], task_output["format"]))

            if "script_data" in outputs[name]:
                del outputs[name]["script_data"]

        girder_worker.events.trigger('run.after', info)

        return outputs
    finally:
        girder_worker.events.trigger('run.finally', info)



from girder_worker.plugins.vtk import dict_to_vtkarrays, dict_to_vtkrow
import vtk

output = vtk.vtkTable()
if len(input["rows"]) > 0:
    dict_to_vtkarrays(input["rows"][0], input["fields"], output.GetRowData())
    for d in input["rows"]:
        dict_to_vtkrow(d, output.GetRowData())



import re
import girder_worker.utils
import tempfile


def _expand_args(args, inputs, taskInputs, tmpDir):
    """
    Expands arguments to the container execution if they reference input
    data. For example, if an input has id=foo, then a container arg of the form
    $input{foo} would be expanded to the runtime value of that input.
    """
    newArgs = []
    regex = re.compile(r'\$input\{([^}]+)\}')

    for arg in args:
        for inputId in re.findall(regex, arg):
            if inputId in inputs:
                arg = arg.replace('$input{%s}' % inputId,
                                  inputs[inputId]['script_data'])
            elif inputId == '_tempdir':
                arg = arg.replace('$input{_tempdir}', tmpDir)

        newArgs.append(arg)

    return newArgs


def run(task, inputs, outputs, task_inputs, task_outputs, **kwargs):
    script = task['script']
    script_fname = tempfile.mktemp()
    with open(script_fname, 'w') as script_file:
        script_file.write(script)

    tmpDir = kwargs.get('_tempdir')
    args = _expand_args(task['swift_args'], inputs, task_inputs, tmpDir)

    print_stderr, print_stdout = True, True
    for id, to in task_outputs.iteritems():
        if id == '_stderr':
            outputs['_stderr']['script_data'] = ''
            print_stderr = False
        elif id == '_stdout':
            outputs['_stdout']['script_data'] = ''
            print_stdout = False

    command = ['swift', script_fname] + args

    print('Running swift: "%s"' % ' '.join(command))

    p = girder_worker.utils.run_process(command, outputs,
                                        print_stdout, print_stderr)

    if p.returncode != 0:
        raise Exception('Error: swift run returned code {}.'.format(
                        p.returncode))

    for name, task_output in task_outputs.iteritems():
        with open(name) as output_file:
            outputs[name]['script_data'] = output_file.read()



import vtk
import networkx as nx
from girder_worker.plugins.vtk import vtkrow_to_dict

directed = isinstance(input, vtk.vtkMutableDirectedGraph)
output = nx.DiGraph() if directed else nx.Graph()

# Add nodes
for node in range(input.GetNumberOfVertices()):
    output.add_node(node, vtkrow_to_dict(input.GetVertexData(), node))

# Add edges
for edge in range(input.GetNumberOfEdges()):
    output.add_edge(input.GetSourceVertex(edge),
                    input.GetTargetVertex(edge),
                    attr_dict=vtkrow_to_dict(input.GetEdgeData(), edge))



import girder_worker.utils
import os
import json


def _write_scala_script(script, inputs, task_outputs, tmp_dir):
    script_fname = os.path.join(tmp_dir, 'script.scala')
    with open(script_fname, 'w') as script_file:
        # Send input values to the script
        for name, binding in inputs.iteritems():
            value = json.dumps(binding['script_data'])
            script_file.write('val ' + name + ' = ' + value + '\n')

        # Run the script
        script_file.write(script)

        # Write output values to temporary files
        script_file.write('import java.io._\n')
        for name in task_outputs:
            if name != '_stderr' and name != '_stdout':
                fname = os.path.join(tmp_dir, name)
                script_file.write("""
new PrintWriter({}) {{
    write({}); close
}}
""".format(json.dumps(fname), name + '.toString()'))

        # Exit the interactive shell
        script_file.write('System.exit(0)\n')

    return script_fname


def _run(spark, task, inputs, outputs, task_inputs, task_outputs, **kwargs):
    tmp_dir = kwargs.get('_tempdir')

    script_fname = _write_scala_script(task['script'], inputs, task_outputs, tmp_dir)

    print_stderr, print_stdout = True, True
    for id, to in task_outputs.iteritems():
        if id == '_stderr':
            outputs['_stderr']['script_data'] = ''
            print_stderr = False
        elif id == '_stdout':
            outputs['_stdout']['script_data'] = ''
            print_stdout = False

    if spark:
        command = ['spark-shell', '-i', script_fname]
    else:
        command = ['scala', script_fname]

    print('Running scala: "%s"' % ' '.join(command))

    p = girder_worker.utils.run_process(command, outputs,
                                        print_stdout, print_stderr)

    if p.returncode != 0:
        raise Exception('Error: scala run returned code {}.'.format(
                        p.returncode))

    for name, task_output in task_outputs.iteritems():
        if name != '_stderr' and name != '_stdout':
            fname = os.path.join(tmp_dir, name)
            with open(fname) as output_file:
                outputs[name]['script_data'] = output_file.read()

            # Deal with converting from string - assume JSON
            if task_output['type'] != 'string':
                outputs[name]['script_data'] = json.loads(outputs[name]['script_data'])


def run(**kwargs):
    _run(spark=False, **kwargs)


def run_spark(**kwargs):
    _run(spark=True, **kwargs)



from girder_worker.plugins.vtk import vtkrow_to_dict

output = {"fields": [], "rows": []}
for c in range(input.GetNumberOfColumns()):
    output["fields"].append(input.GetColumnName(c))
for r in range(input.GetNumberOfRows()):
    output["rows"].append(vtkrow_to_dict(input.GetRowData(), r))



import os
import girder_worker


def vtkrow_to_dict(attributes, i):
    row = {}
    for c in range(attributes.GetNumberOfArrays()):
        arr = attributes.GetAbstractArray(c)
        values = []
        comp = arr.GetNumberOfComponents()
        for c in range(comp):
            variant_value = arr.GetVariantValue(i*comp + c)
            if variant_value.IsInt():
                value = variant_value.ToInt()
            elif variant_value.IsLong():
                value = variant_value.ToLong()
            elif variant_value.IsDouble() or variant_value.IsFloat():
                value = variant_value.ToDouble()
            else:
                value = variant_value.ToString()
            values.append(value)
        row[arr.GetName()] = values[0] if len(values) == 1 else values
    return row


def dict_to_vtkarrays(row, fields, attributes):
    import vtk
    for key in fields:
        value = row[key]
        comp = 1
        if isinstance(value, list):
            comp = len(value)
            value = value[0]
        if isinstance(value, (int, long, float)):
            arr = vtk.vtkDoubleArray()
        elif isinstance(value, str):
            arr = vtk.vtkStringArray()
        elif isinstance(value, unicode):
            arr = vtk.vtkUnicodeStringArray()
        else:
            arr = vtk.vtkStringArray()
        arr.SetName(key)
        arr.SetNumberOfComponents(comp)
        attributes.AddArray(arr)


def dict_to_vtkrow(row, attributes):
    for key in row:
        value = row[key]
        if not isinstance(value, (list, int, long, float, str, unicode)):
            value = str(value)
        found = False
        for i in range(attributes.GetNumberOfArrays()):
            arr = attributes.GetAbstractArray(i)
            if arr.GetName() == key:
                if isinstance(value, list):
                    for v in value:
                        arr.InsertNextValue(v)
                else:
                    arr.InsertNextValue(value)
                found = True
                break
        if not found:
            raise Exception("[dict_to_vtkrow] Unexpected key: " + key)


def load(params):
    converters_dir = os.path.join(params['plugin_dir'], 'converters')
    girder_worker.format.import_converters([
        os.path.join(converters_dir, 'geometry'),
        os.path.join(converters_dir, 'graph'),
        os.path.join(converters_dir, 'table'),
        os.path.join(converters_dir, 'tree')
    ])



from girder_worker.plugins.vtk import dict_to_vtkarrays, dict_to_vtkrow
import vtk

vtk_builder = vtk.vtkMutableDirectedGraph()
node_fields = input["node_fields"]
edge_fields = input["edge_fields"]
dict_to_vtkarrays(input["node_data"], node_fields, vtk_builder.GetVertexData())
if "children" in input and len(input["children"]) > 0:
    if "edge_data" in input["children"][0]:
        dict_to_vtkarrays(input["children"][0]["edge_data"], edge_fields,
                          vtk_builder.GetEdgeData())


def process_node(vtknode, node):
    if "children" in node:
        for n in node["children"]:
            vtkchild = vtk_builder.AddVertex()
            vtk_builder.AddGraphEdge(vtknode, vtkchild).GetId()
            dict_to_vtkrow(n["node_data"], vtk_builder.GetVertexData())
            if "edge_data" in n:
                dict_to_vtkrow(n["edge_data"], vtk_builder.GetEdgeData())
            process_node(vtkchild, n)
vtk_builder.AddVertex()
dict_to_vtkrow(input["node_data"], vtk_builder.GetVertexData())
process_node(0, input)
output = vtk.vtkTree()
output.ShallowCopy(vtk_builder)



import os
import re
import requests


def _readFilenameFromResponse(request, url):
    """
    This helper will derive a filename from the HTTP response, first attempting
    to use the content disposition header, otherwise falling back to the last
    token of the URL.
    """
    match = re.search('filename="(.*)"',
                      request.headers.get('Content-Disposition', ''))

    if match is None:
        return [t for t in url.split('/') if t][-1]
    else:
        return match.group(1)


def fetch(spec, **kwargs):
    """
    Downloads an input file via HTTP using requests.
    """
    if 'url' not in spec:
        raise Exception('No URL specified for HTTP input.')
    taskInput = kwargs.get('task_input', {})
    target = taskInput.get('target', 'memory')
    url = spec['url']
    method = spec.get('method', 'GET').upper()
    request = requests.request(method, url, headers=spec.get('headers', {}),
                               stream=True, allow_redirects=True)

    try:
        request.raise_for_status()
    except Exception:
        print 'HTTP fetch failed (%s). Response: %s' % (url, request.text)
        raise

    if target == 'filepath':
        tmpDir = kwargs['_tempdir']

        if 'filename' in taskInput:
            filename = taskInput['filename']
        else:
            filename = _readFilenameFromResponse(request, url)

        path = os.path.join(tmpDir, filename)

        total = 0
        maxSize = spec.get('maxSize')

        with open(path, 'wb') as out:
            for buf in request.iter_content(65536):
                length = len(buf)
                if maxSize and length + total > maxSize:
                    raise Exception(
                        'Exceeded max download size of %d bytes.' % maxSize)
                out.write(buf)
                total += length

        return path
    elif target == 'memory':
        return ''.join(request.iter_content(65536))
    else:
        raise Exception('Invalid HTTP fetch target: ' + target)


def push(data, spec, **kwargs):
    taskOutput = kwargs.get('task_output', {})
    target = taskOutput.get('target', 'memory')

    url = spec['url']
    method = spec.get('method', 'POST').upper()

    if target == 'filepath':
        with open(data, 'rb') as fd:
            request = requests.request(
                method, url, headers=spec.get('headers', {}), data=fd,
                allow_redirects=True)
    elif target == 'memory':
        request = requests.request(
            method, url, headers=spec.get('headers', {}), data=data,
            allow_redirects=True)
    else:
        raise Exception('Invalid HTTP fetch target: ' + target)

    try:
        request.raise_for_status()
    except Exception:
        print 'HTTP push failed (%s). Response: %s' % (url, request.text)
        raise



import os
import re
import girder_worker.utils
import subprocess


def _pull_image(image):
    """
    Pulls the specified docker image onto this worker.
    """
    command = ('docker', 'pull', image)
    p = subprocess.Popen(args=command, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()

    if p.returncode != 0:
        print('Error pulling docker image %s:' % image)
        print('STDOUT: ' + stdout)
        print('STDERR: ' + stderr)

        raise Exception('Docker pull returned code {}.'.format(p.returncode))


def _transform_path(inputs, taskInputs, inputId, tmpDir):
    """
    If the input specified by inputId is a filepath target, we transform it to
    its absolute path within the docker container (underneath /data).
    """
    for ti in taskInputs.itervalues():
        tiId = ti['id'] if 'id' in ti else ti['name']
        if tiId == inputId:
            if ti.get('target') == 'filepath':
                rel = os.path.relpath(inputs[inputId]['script_data'], tmpDir)
                return os.path.join('/data', rel)
            else:
                return inputs[inputId]['script_data']

    raise Exception('No task input found with id = ' + inputId)


def _expand_args(args, inputs, taskInputs, tmpDir):
    """
    Expands arguments to the container execution if they reference input
    data. For example, if an input has id=foo, then a container arg of the form
    $input{foo} would be expanded to the runtime value of that input. If that
    input is a filepath target, the file path will be transformed into the
    location that it will be available inside the running container.
    """
    newArgs = []
    regex = re.compile(r'\$input\{([^}]+)\}')

    for arg in args:
        for inputId in re.findall(regex, arg):
            if inputId in inputs:
                transformed = _transform_path(inputs, taskInputs, inputId,
                                              tmpDir)
                arg = arg.replace('$input{%s}' % inputId, transformed)
            elif inputId == '_tempdir':
                arg = arg.replace('$input{_tempdir}', '/data')

        newArgs.append(arg)

    return newArgs


def run(task, inputs, outputs, task_inputs, task_outputs, **kwargs):
    image = task['docker_image']
    print('Pulling docker image: ' + image)
    _pull_image(image)

    tmpDir = kwargs.get('_tempdir')
    args = _expand_args(task['container_args'], inputs, task_inputs, tmpDir)

    print_stderr, print_stdout = True, True
    for id, to in task_outputs.iteritems():
        if id == '_stderr':
            outputs['_stderr']['script_data'] = ''
            print_stderr = False
        elif id == '_stdout':
            outputs['_stdout']['script_data'] = ''
            print_stdout = False

    command = ['docker', 'run', '--rm', '-u', str(os.getuid())]

    if tmpDir:
        command += ['-v', tmpDir + ':/data']

    if 'entrypoint' in task:
        command += ['--entrypoint', task['entrypoint']]

    command += [image] + args

    print('Running container: "%s"' % ' '.join(command))

    p = girder_worker.utils.run_process(command, outputs,
                                        print_stdout, print_stderr)

    if p.returncode != 0:
        raise Exception('Error: docker run returned code {}.'.format(
                        p.returncode))

    for name, task_output in task_outputs.iteritems():
        # TODO grab files written inside the container somehow?
        pass



import girder_worker

from girder_worker.utils import toposort


def run(task, inputs, outputs, task_inputs, task_outputs, validate,  # noqa
        auto_convert, **kwargs):
    # Make map of steps
    steps = {step["name"]: step for step in task["steps"]}

    # Make map of input bindings
    bindings = {step["name"]: {} for step in task["steps"]}

    # Create dependency graph and downstream pointers
    dependencies = {step["name"]: set() for step in task["steps"]}
    downstream = {}
    for conn in task["connections"]:
        # Add dependency graph link for internal links
        if "input_step" in conn and "output_step" in conn:
            dependencies[conn["input_step"]].add(conn["output_step"])

        # Add downstream links for links with output
        if "output_step" in conn:
            ds = downstream.setdefault(conn["output_step"], {})
            ds_list = ds.setdefault(conn["output"], [])
            ds_list.append(conn)

        # Set initial bindings for inputs
        if "input_step" in conn and "output_step" not in conn:
            name = conn["name"]
            bindings[conn["input_step"]][conn["input"]] = {
                "format": task_inputs[name]["format"],
                "data": inputs[name]["script_data"]
            }

    # Traverse analyses in topological order
    for step_set in toposort(dependencies):
        for step in step_set:
            # Visualizations cannot be executed
            if ("visualization" in steps[step] and
                    steps[step]["visualization"]):
                continue

            # Run step
            print "--- beginning: %s ---" % steps[step]["name"]
            out = girder_worker.run(steps[step]["task"], bindings[step])
            print "--- finished: %s ---" % steps[step]["name"]

            # Update bindings of downstream analyses
            if step in downstream:
                for name, conn_list in downstream[step].iteritems():
                    for conn in conn_list:
                        if "input_step" in conn:
                            # This is a connection to a downstream step
                            b = bindings[conn["input_step"]]
                            b[conn["input"]] = out[name]
                        else:
                            # This is a connection to a final output
                            o = outputs[conn["name"]]
                            o["script_data"] = out[name]["data"]

    # Output visualization paramaters
    outputs["_visualizations"] = []
    for step in task["steps"]:
        if "visualization" not in step or not step["visualization"]:
            continue
        vis_bindings = {}
        for b, value in bindings[step["name"]].iteritems():
            script_output = value
            vis_input = None
            for step_input in step["task"]["inputs"]:
                if step_input["name"] == b:
                    vis_input = step_input

            if not vis_input:
                raise Exception(
                    "Could not find visualization input named " + b + "."
                )

            # Validate the output
            if (validate and not
                    girder_worker.isvalid(vis_input["type"], script_output)):
                raise Exception(
                    "Output %s (%s) is not in the expected type (%s) and "
                    "format (%s)." % (name, type(script_output["data"]),
                                      vis_input["type"], script_output["format"]))

            if auto_convert:
                vis_bindings[b] = girder_worker.convert(
                    vis_input["type"],
                    script_output,
                    {"format": vis_input["format"]}
                )

            elif script_output["format"] == vis_input["format"]:
                data = script_output["data"]
                if "mode" in script_output:
                    girder_worker.io.push(data, script_output)
                else:
                    vis_bindings[b] = {
                        "type": vis_input["type"],
                        "format": vis_input["format"],
                        "data": data
                    }
            else:
                raise Exception(
                    "Expected exact format match but '" +
                    script_output["format"] +
                    "' != '" + vis_input["format"] + "'."
                )

            if "script_data" in vis_bindings[b]:
                del vis_bindings[b]["script_data"]

        outputs["_visualizations"].append({
            "mode": "preset",
            "type": step["name"],
            "inputs": vis_bindings
        })



import girder_client
import os
import girder_worker

from six import StringIO


def _init_client(spec, require_token=False):
    if 'api_url' in spec:
        client = girder_client.GirderClient(apiUrl=spec['api_url'])
    elif 'host' in spec:
        scheme = spec.get('scheme', 'http')
        port = spec.get('port', {
            'http': 80,
            'https': 443
        }[scheme])
        api_root = spec.get('api_root', '/api/v1')
        client = girder_client.GirderClient(
            host=spec['host'], scheme=scheme, apiRoot=api_root, port=port)

    if 'token' in spec:
        client.token = spec['token']
    elif require_token:
        raise Exception('You must pass a token for Girder authentication.')
    else:
        raise Exception('You must pass either an api_url or host key for '
                        'Girder input and output bindings.')

    return client


def fetch_handler(spec, **kwargs):
    resource_type = spec.get('resource_type', 'file').lower()
    taskInput = kwargs.get('task_input', {})
    target = taskInput.get('target', 'filepath')

    if 'id' not in spec:
        raise Exception('Must pass a resource ID for girder inputs.')
    if 'name' not in spec:
        raise Exception('Must pass a name for girder inputs.')

    client = _init_client(spec)
    dest = os.path.join(kwargs['_tempdir'], spec['name'])

    if resource_type == 'folder':
        client.downloadFolderRecursive(spec['id'], dest)
    elif resource_type == 'item':
        client.downloadItem(spec['id'], kwargs['_tempdir'], spec['name'])
    elif resource_type == 'file':
        client.downloadFile(spec['id'], dest)
    else:
        raise Exception('Invalid resource type: ' + resource_type)

    if target == 'filepath':
        return dest
    elif target == 'memory':
        with open(dest, 'rb') as fd:
            return fd.read()
    else:
        raise Exception('Invalid Girder push target: ' + target)


def push_handler(data, spec, **kwargs):
    parent_type = spec.get('parent_type', 'folder')
    taskOutput = kwargs.get('task_output', {})
    target = taskOutput.get('target', 'filepath')

    if 'parent_id' not in spec:
        raise Exception('Must pass parent ID for girder outputs.')

    client = _init_client(spec, require_token=True)

    if target == 'memory':
        if not spec.get('name'):
            raise Exception('Girder uploads from memory objects must '
                            'explicitly pass a "name" field.')
        fd = StringIO(data)
        client.uploadFile(parentId=spec['parent_id'], stream=fd, size=len(data),
                          parentType=parent_type, name=spec['name'])
    elif target == 'filepath':
        name = spec.get('name') or os.path.basename(data)
        size = os.path.getsize(data)
        with open(data, 'rb') as fd:
            client.uploadFile(parentId=spec['parent_id'], stream=fd, size=size,
                              parentType=parent_type, name=name)
    else:
        raise Exception('Invalid Girder push target: ' + target)


def load(params):
    girder_worker.io.register_fetch_handler('girder', fetch_handler)
    girder_worker.io.register_push_handler('girder', push_handler)



from girder_worker.plugins.vtk import dict_to_vtkarrays, dict_to_vtkrow
import six
import vtk

"""
NetworkX to vtkGraph

This creates a vtkMutableDirectedGraph or vtkMutableUndirectedGraph.

It creates nodes which lose their actual value since nodes in vtkGraph are referenced
strictly by their index. These nodes do retain metadata in the same way edges do
(see below).

It creates edges which maintain the proper association and any metadata, the caveat
is that all edges are given all metadata attributes with default values.

So if an edge has an integer 'distance' attribute, and another does not - the
non-distanced edge will have a distance of 0. This follows suit for all of pythons
defaults, bool(), str(), float(), etc.

As such it requires that all keys have the same type.
"""

if input.is_directed():
    output = vtk.vtkMutableDirectedGraph()
else:
    output = vtk.vtkMutableUndirectedGraph()

nodes = input.nodes(data=True)
edges = input.edges(data=True)
edge_field_types = {}
node_field_types = {}

# Find out fields and types
for (_, data) in nodes:
    for key in data.keys():
        data_type = type(data[key])

        # Assert that every node which has key has the same type
        if key in node_field_types and data_type != node_field_types[key]:
            raise Exception('Node has heterogeneous types for key %s' % key)

        node_field_types[key] = data_type

for (_, _, data) in edges:
    for key in data.keys():
        data_type = type(data[key])

        # Assert that every edge which has key has the same type
        if key in edge_field_types and data_type != edge_field_types[key]:
            raise Exception('Edge has heterogeneous types for key %s' % key)

        edge_field_types[key] = data_type

# Merge default attributes into nodes and edges
for (_, data) in nodes:
    for (field, field_type) in six.iteritems(node_field_types):
        if field not in data:
            data[field] = field_type()

for (_, _, data) in edges:
    for (field, field_type) in six.iteritems(edge_field_types):
        if field not in data:
            data[field] = field_type()

# Add vtkArrays to the output data for nodes and edges
# We can just use the first node and edge since they all have same attr name/types
if nodes:
    dict_to_vtkarrays(nodes[0][1],
                      node_field_types.keys(),
                      output.GetVertexData())

if edges:
    dict_to_vtkarrays(edges[0][2],
                      edge_field_types.keys(),
                      output.GetEdgeData())


# This is a mapping of NetworkX nodes to VTK Vertex IDs so we can refer to them
# when adding edges between NetworkX nodes later
node_to_ids = {}

for (node, data) in nodes:
    node_to_ids[node] = output.AddVertex()
    dict_to_vtkrow(data, output.GetVertexData())

for (u, v, data) in edges:
    output.AddGraphEdge(node_to_ids[u], node_to_ids[v])
    dict_to_vtkrow(data, output.GetEdgeData())



import girder_worker
from girder_worker.format import conv_graph

from .utils import JobManager, JobStatus
from celery import Celery

app = None


def main():
    global app
    app = Celery(
        main=girder_worker.config.get('celery', 'app_main'),
        backend=girder_worker.config.get('celery', 'broker'),
        broker=girder_worker.config.get('celery', 'broker'))

    @app.task
    def run(*pargs, **kwargs):
        jobInfo = kwargs.pop('jobInfo', {})
        retval = 0

        with JobManager(logPrint=jobInfo.get('logPrint', True),
                        url=jobInfo.get('url'), method=jobInfo.get('method'),
                        headers=jobInfo.get('headers')) as jm:
            kwargs['_job_manager'] = jm
            kwargs['status'] = JobStatus.RUNNING
            retval = girder_worker.run(*pargs, **kwargs)
        return retval

    @app.task
    def convert(*pargs, **kwargs):
        return girder_worker.convert(*pargs, **kwargs)

    @app.task
    def validators(*pargs, **kwargs):
        _type, _format = pargs
        nodes = []

        for (node, data) in conv_graph.nodes(data=True):
            if ((_type is None) or (_type == node.type)) and \
               ((_format is None) or (_format == node.format)):
                nodes.append({'type': node.type,
                              'format': node.format,
                              'validator': data})

        return nodes

    app.worker_main()


if __name__ == '__main__':
    main()



import girder_worker
import os
import sys

from ConfigParser import NoOptionError, NoSectionError


def setup_spark_env():
    # Setup pyspark
    spark_home = None

    try:
        spark_home = girder_worker.config.get('spark', 'spark_home')
    except (NoOptionError, NoSectionError):
        pass

    # If not configured try the environment
    if not spark_home:
        spark_home = os.environ.get('SPARK_HOME')

    if not spark_home:
        raise Exception('spark_home must be set or SPARK_HOME must be set in '
                        'the environment')

    # Need to set SPARK_HOME
    os.environ['SPARK_HOME'] = spark_home

    if not os.path.exists(spark_home):
        raise Exception('spark_home is not a valid directory')

    sys.path.append(os.path.join(spark_home, 'python'))
    sys.path.append(os.path.join(spark_home, 'bin'))

    # Check that we can import SparkContext
    from pyspark import SparkConf, SparkContext  # noqa


def create_spark_context(task_spark_conf):
    from pyspark import SparkConf, SparkContext
    # Set can spark configuration parameter user has specified
    spark_conf = SparkConf()

    if girder_worker.config.has_section('spark'):
        for (name, value) in girder_worker.config.items('spark'):
            spark_conf.set(name, value)

    # Override with any task specific configuration
    for (name, value) in task_spark_conf.items():
        spark_conf.set(name, value)

    # Build up the context, using the master URL
    sc = SparkContext(conf=spark_conf)

    return sc



import os
import girder_worker
from . import pyspark_executor, spark

SC_KEY = '_girder_worker_spark_context'


def setup_pyspark_task(event):
    """
    This is executed before a task execution. If it is a pyspark task, we
    create the spark context here so it can be used for any input conversion.
    """
    info = event.info
    if info['mode'] == 'spark.python' and SC_KEY not in info['kwargs']:
        spark_conf = info['task'].get('spark_conf', {})
        info['kwargs'][SC_KEY] = spark.create_spark_context(spark_conf)
        info['cleanup_spark'] = True


def pyspark_run_cleanup(event):
    if event.info.get('cleanup_spark'):
        event.info['kwargs'][SC_KEY].stop()


def load(params):
    # If we have a spark config section then try to setup spark environment
    if girder_worker.config.has_section('spark') or 'SPARK_HOME' in os.environ:
        spark.setup_spark_env()

    girder_worker.register_executor('spark.python', pyspark_executor.run)

    girder_worker.events.bind('run.before', 'spark', setup_pyspark_task)
    girder_worker.events.bind('run.finally', 'spark', pyspark_run_cleanup)

    girder_worker.format.import_converters(
        os.path.join(params['plugin_dir'], 'converters'))



import rpy2.robjects


def run(task, inputs, outputs, task_inputs, task_outputs, **kwargs):
    env = rpy2.robjects.globalenv

    # Clear out workspace variables and packages
    rpy2.robjects.reval("""
        rm(list = ls())
        pkgs <- names(sessionInfo()$otherPkgs)
        if (!is.null(pkgs)) {
            pkgs <- paste('package:', pkgs, sep = "")
            lapply(pkgs, detach, character.only = TRUE, unload = TRUE)
        }
        """, env)

    env["tempdir"] = kwargs.get("_tempdir")

    for name in inputs:
        env[str(name)] = inputs[name]["script_data"]

    rpy2.robjects.reval(task["script"], env)

    for name, task_output in task_outputs.iteritems():
        d = outputs[name]
        d["script_data"] = env[str(name)]

        # Hack to detect scalar values from R.
        # The R value might not have a len() so wrap in a try/except.
        try:
            if len(d["script_data"]) == 1:
                d["script_data"] = d["script_data"][0]
        except TypeError:
            pass
